package cologne.eck.all_peas.gui;

/*
 * PeaFactory - Production of Password Encryption Archives
 * Copyright (C) 2021  Axel von dem Bruch
 * 
 * This library is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published 
 * by the Free Software Foundation; either version 2 of the License, 
 * or (at your option) any later version.
 * This library is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * See:  http://www.gnu.org/licenses/gpl-2.0.html
 * You should have received a copy of the GNU General Public License 
 * along with this library.
 */

import java.io.File;
import java.io.IOException;
import java.util.Random;


import org.bouncycastle.crypto.digests.Blake2bDigest;

import cologne.eck.all_peas.data.AttachmentHandler;
import cologne.eck.all_peas.data.PeaProperties;
import cologne.eck.all_peas.files.FileComposer;
import cologne.eck.all_peas.files.FileModel;
import cologne.eck.all_peas.files.FileTypePanel;
import cologne.eck.all_peas.vm_specific.JREBridge;
import cologne.eck.all_peas.vm_specific.JREProperties;
import cologne.eck.peafactory.crypto.CipherStuff;
import cologne.eck.peafactory.crypto.HashStuff;
import cologne.eck.peafactory.crypto.KeyFileHandler;
import cologne.eck.peafactory.crypto.kdf.KeyDerivation;
import cologne.eck.tools.Converter;
import cologne.eck.tools.EntropyPool;
import cologne.eck.tools.PropertyHandler;
import cologne.eck.tools.TestLog;
import cologne.eck.tools.Zeroizer;
import cologne.eck.tools.ZipStuff;

public class KeyFileDialogs {
	
	private static String probablyNewDefaultKeyFileName = null;
	
	private static Random random = new Random();
	
	/**
	 * For manual tests....
	 * @param args
	 */
	public static void main(String[] args) {
		
		PeaProperties.setVmBridge(new JREBridge());
		JREProperties.setTypePanel(new FileTypePanel(0, 0, true, false, 
				new FileComposer(
						new FileModel(true, AttachmentHandler.getVersion2ID(), new byte[KeyDerivation.getSaltSize()])), args));
		HashStuff.setHashAlgo(new Blake2bDigest());
		EntropyPool pool = EntropyPool.getInstance();
		pool.updateThread(1);
		
		
		int createSuccess = setKeyFileStatus(true);
		TestLog.o(KeyFileDialogs.class, "Created key file: " + createSuccess);
		
//		File f = manuallyChooseNewKeyFile();
//		TestLog.o(KeyFileDialogs.class, "File: " + f.getAbsolutePath());
	}
	/**
	 * Apply the key file (default or user chosen)
	 * 
	 * @param keyMaterial	the key material from KDF
	 * 
	 * @return	hash of keyMaterial and key file hash (HKDF)
	 */
	public static byte[] applyKeyFile(byte[] keyMaterial) {
		
		if (keyMaterial == null) {
			TestLog.e(KeyFileDialogs.class, "Missing keyMaterial...");		
			TestLog.ve(KeyFileDialogs.class, "Missing keyMaterial...", 5);		
			PeaDialog.showMessage(JREProperties.getMainWindow(), 
							PeaProperties.getVmBridge().translate("unexpected_error"), 
							PeaProperties.getVmBridge().translate("error"), 0); 
			return null;
		}
		
		byte[] keyFileHash = null;
		int keyFileFlag = -1;
		
		if (JREProperties.getMainWindow() instanceof PswDialogView) {
			// default
			keyFileFlag = PeaProperties.getKeyFileFlag();			
		} else if (JREProperties.getMainWindow() instanceof PeaLockFrame) {
			// open file
			keyFileFlag = PeaProperties.getKeyFileFlag();
		} else {
			TestLog.e(KeyFileDialogs.class, "Invalid main window");
			TestLog.ve(KeyFileDialogs.class, "Invalid main window", 5);		
			PeaDialog.showMessage(JREProperties.getMainWindow(), 
							PeaProperties.getVmBridge().translate("unexpected_error"), 
							PeaProperties.getVmBridge().translate("error"), 0);
			return null;
		}
		
		TestLog.v(KeyFileDialogs.class, "keyFileFlag:" + keyFileFlag);	
		
		// Get the key file hash
		if (keyFileFlag == 0) {
			return keyMaterial;
		} else if (keyFileFlag == 1) {
			// use default
			keyFileHash = getDefaultKeyFileHash();
		} else if (keyFileFlag == 2) {
			// choose manually
			keyFileHash = getChosenKeyFileHash();
		}
		
		if (keyFileHash == null) { // probably cancelled user chosen key file
			TestLog.e(KeyFileDialogs.class, "Missing key file hash");		
	/*		PeaDialog.showMessage(JREProperties.getMainWindow(), 
							PeaProperties.getVmBridge().translate("unexpected_error"), 
							PeaProperties.getVmBridge().translate("error"), 0); */
			return null;
		}
		
		// Use HKDF 
		byte[] result =  HashStuff.generateHKDFBytes(keyMaterial, 
				keyFileHash, 
				CipherStuff.getKeySize() ); // size 
		Zeroizer.zero(keyMaterial);
		Zeroizer.zero(keyFileHash);
		return result;
	}

	/**
	 * Ask for kind of key file, 
	 * Create or choose a new or no key file. 
	 * If this is a default key file, store in properties. 
	 * 
	 * @param changeDefault	true: only default key file, false: manual key file allowed
	 * 
	 * @return	-1 for cancel or error, 0 for no key file, 1 for default, 2 for manual chosen
	 */
	public static int setKeyFileStatus(boolean changeDefault) {
		
		String[] options = { 
				PeaProperties.getVmBridge().translate("gen_key_file_automatically_opened"),
				PeaProperties.getVmBridge().translate("no_key_file"),
				PeaProperties.getVmBridge().translate("gen_key_file_manually_opened"),
				PeaProperties.getVmBridge().translate("choose_key_file_automatically_opened")};
				//	"Generate default key file to be opened automatically (recommanded)", 
				//	"No key file",			
				//	"Generate default key file to be selected manually each time", 
				//	"Choose any file to be selected manually each time" };

		int chosenType = PeaDialog.showOptionYAxis(JREProperties.getMainWindow(), 
				PeaProperties.getVmBridge().translate("choose_key_file_kind") + "\n" +
				PeaProperties.getVmBridge().translate("default_key_file_info"),
		//		"Choose the kind of your key file. \n"
		//				+ "A default key file is opened automatically.", 
						PeaProperties.getVmBridge().translate("key_file"), 3, options, 0);
		TestLog.o(KeyFileDialogs.class, "Chosen option: " + chosenType);		
		
		boolean createKeyFile = false;
		int keyFileFlag = -1;
		if (chosenType == -1 ) { // cancel
			return -1;
		} else if (chosenType == 1) { // no key file
			keyFileFlag = 0;
			createKeyFile = false;
		} else if (chosenType == 0) { // generate default
			keyFileFlag = 1;
			createKeyFile = true;
		} else if (chosenType == 2 ) { // select manually
			keyFileFlag = 2;
			createKeyFile = true;
		} else if (chosenType == 3 ) { // select manually
			keyFileFlag = 2;
			createKeyFile = false;
		} else {
			return -1;
		}
		
		// Set the flags and properties for no key file
		if (keyFileFlag == 0) {// no key file
			if (changeDefault == true) {
				PeaProperties.setDefaultKeyFileFlag(0); // default 
				PeaProperties.setDefaultKeyFileName(null);
			} else {
				PeaProperties.setKeyFileFlag(0); // actual 
			}
			return 0;
		}
		
		File keyFile = null;
		if (createKeyFile == true) {
			keyFile = generateKeyFile();
			if (keyFile == null) {
				return -1;
			}

		} else {
			
			keyFile = manuallyChooseNewKeyFile();
			if (keyFile == null) {			
				return -1;
			}
		}
		
		// Set key file flags and properties:
		if (keyFileFlag == 1) {
			PeaProperties.setDefaultKeyFileName(keyFile.getAbsolutePath());
			// store in pea.properties
			String propError = new PropertyHandler().setOneProperty("key_file", keyFile.getAbsolutePath());
			if (propError != null) {
				// no property file e.g. first start -> set file
				new PropertyHandler().updateAllPropValues();
				propError = new PropertyHandler().setOneProperty("key_file", keyFile.getAbsolutePath());
				if (propError != null) {
					TestLog.ve(KeyFileDialogs.class, "Could not store default key file in poperties: " + propError, 5);
					PeaDialog.showMessage(JREProperties.getMainWindow(), 
							PeaProperties.getVmBridge().translate("operation_failed") + "\n"
									+ propError + "\n\n"
									+ PeaProperties.getVmBridge().translate("select_file_when_open_files"), 
									PeaProperties.getVmBridge().translate("unexpected_error"), 0);
				}
			}
			PeaDialog.showMessage(JREProperties.getMainWindow(), 					
					PeaProperties.getVmBridge().translate("key_file_must_be_present") + ": \n\n" 
			+ keyFile.getAbsolutePath() + "\n\n"
			+ PeaProperties.getVmBridge().translate("never_modify_or_delete") ); 
			
			if (changeDefault == true) {
				PeaProperties.setDefaultKeyFileFlag(1); // default 
			} else {
				PeaProperties.setKeyFileFlag(1); //actual file
			}

		} else if (keyFileFlag == 2) {
			PeaDialog.showMessage(JREProperties.getMainWindow(), 					
					PeaProperties.getVmBridge().translate("select_file_when_open_files") + ": \n\n" 
							+ keyFile.getAbsolutePath()  + "\n\n"
							+ PeaProperties.getVmBridge().translate("never_modify_or_delete") ); 
			// manual selection
			if (changeDefault == true) {
				PeaProperties.setDefaultKeyFileFlag(2); // default 
			} else {
				PeaProperties.setKeyFileFlag(2); //actual file
			}
		}  else {
			return -1;
		}
		
		return keyFileFlag;
	}
	
	public static File generateKeyFile() {
		int randomInt = Math.abs( random.nextInt() );
		String randomString =  Converter.bytes2hex( Converter.int2bytesBE(randomInt) ); //Integer.toString(randomInt);
		// 2. Select file name and directory 
		PeaFileChooser pfc = new PeaFileChooser(JREProperties.getMainWindow());
		pfc.customizeFileChooser(null, 
				PeaProperties.getVmBridge().translate("select_folder_for_key_file"), 
				true);
		pfc.setFileSelectionMode(PeaFileChooser.FILES_ONLY);
		pfc.setSelectedFile(new File(randomString ));
		pfc.showOpenDialog(JREProperties.getMainWindow());
		File keyFile = pfc.getSelectedFile();	
		if (keyFile == null) {			
			return null;
		}
		TestLog.v(KeyFileDialogs.class, keyFile.getAbsolutePath());

		if (keyFile.exists()) {
			PeaDialog.showMessage(JREProperties.getMainWindow(), 
					PeaProperties.getVmBridge().translate("file_already_exists")
					+ "\n" + 
					keyFile.getAbsolutePath() +"\n" + 
							PeaProperties.getVmBridge().translate("select_new_file"), 
							PeaProperties.getVmBridge().translate("error"), 0);
			return null;
		}


		// 3. Create empty file
		if (keyFile.getParentFile() != null) {
			keyFile.getParentFile().mkdirs(); 
		}
		try {
			keyFile.createNewFile();
		} catch (IOException e) {
			TestLog.e(KeyFileDialogs.class, e.getLocalizedMessage());
			PeaDialog.showMessage(JREProperties.getMainWindow(), 
					e.getLocalizedMessage(), 
					PeaProperties.getVmBridge().translate("error"), 0);
			return null;
		}
		String error = KeyFileHandler.createKeyFile(keyFile);
		if (error != null) {
			PeaDialog.showMessage(JREProperties.getMainWindow(), 
					PeaProperties.getVmBridge().translate("operation_failed") + "\n" + error, 
					PeaProperties.getVmBridge().translate("error") + ": " + PeaProperties.getVmBridge().translate("key_file"), 0);
			return null;
		}		
		return keyFile;
	}
	
	
	/**
	 * Choose any file as key file. 
	 * This will check the size and the entropy of the file
	 * and warn if this is not appropriate
	 * 
	 * @return	the user chosen key file
	 */
	private static File manuallyChooseNewKeyFile() {
		PeaFileChooser pfc = new PeaFileChooser(JREProperties.getMainWindow());
		pfc.setFileSelectionMode(PeaFileChooser.FILES_ONLY);
		pfc.showOpenDialog(JREProperties.getMainWindow());
		File keyFile = pfc.getSelectedFile();
		
		if (keyFile == null) {
			return null;
		}
		
		long keyFileSize = keyFile.length();
		// Check the file
		if (keyFileSize > 1024 * 1024 * 16) { // 16 MB
			//  warn performance			
			int largeWarning = PeaDialog.showQuestion(JREProperties.getMainWindow(), 
					PeaProperties.getVmBridge().translate("warning") + "\n\n" +
					keyFile.getAbsolutePath() + "\n " + PeaProperties.getVmBridge().translate("size") + ": " + keyFile.length() + "\n "
					+ PeaProperties.getVmBridge().translate("size_warning"), 
					PeaProperties.getVmBridge().translate("warning"), 2);
			if (largeWarning != 0) {
				return null;
			}
		}
		
		if (keyFileSize < 1024) {
			// warn small size
			int smallWarning = PeaDialog.showQuestion(JREProperties.getMainWindow(), 
					PeaProperties.getVmBridge().translate("warning") + "\n\n  " +
					keyFile.getAbsolutePath() + "\n  " + PeaProperties.getVmBridge().translate("size") + ": " + keyFile.length() + "\n"
					+ "  " + PeaProperties.getVmBridge().translate("should_be") + " > " + 1024
					+ "\n\n" + PeaProperties.getVmBridge().translate("continue") + "?",  
					PeaProperties.getVmBridge().translate("warning"), 2);
			if (smallWarning != 0) {
				return null;
			}
		}
		// Zip the file, compare size of zip with origin to get entropy
		String[] fileNames = { keyFile.getAbsolutePath() };
		String tempDir = System.getProperty("java.io.tmpdir");
		long randomLong = Math.abs( random.nextLong() );
		String randomString = Long.toString(randomLong);
		String zipError = ZipStuff.zipFiles(fileNames, randomString, new File(tempDir) );
		if (zipError != null) {
			PeaDialog.showMessage(JREProperties.getMainWindow(), 
					"Couldn't measure entropy of " + keyFile.getAbsolutePath(), 
					PeaProperties.getVmBridge().translate("warning"), 2);
		} else {
			String zipFileName = null;
			
			if (tempDir.endsWith(File.separator)) {
				zipFileName = tempDir + randomString;
			} else {
				zipFileName = tempDir + File.separator + randomString;
			}
			File zipFile = new File(zipFileName);
			if ( ! zipFile.exists()) {
				TestLog.e(KeyFileDialogs.class, "Could not create zip file to measure entropy: " + zipFileName);
				PeaDialog.showMessage(JREProperties.getMainWindow(), 
						"Couldn't measure entropy of " + keyFile.getAbsolutePath(),
						PeaProperties.getVmBridge().translate("warning"), 2);
			} else {
				long zipSize = zipFile.length();
				//TestLog.o(KeyFileDialogs.class, "keyFileSize: " + keyFileSize + ", zipSize: " + zipSize);
				zipFile.delete();
				if ( ((zipSize * 1.5) < keyFileSize) && zipSize < 1024 * 32 ) {
					// warning low entropy
					// warn small size
					int entropyWarning = PeaDialog.showQuestion(JREProperties.getMainWindow(), 
							PeaProperties.getVmBridge().translate("warning") + "\n\n   " +
							keyFile.getAbsolutePath() + ":\n"
									+ "\n   " + PeaProperties.getVmBridge().translate("low_entropy") 
									+ "\n\n" + PeaProperties.getVmBridge().translate("continue") + "?", 
									PeaProperties.getVmBridge().translate("warning"), 2);
					if (entropyWarning != 0) {
						return null;
					}
				}
			}
		}
		return keyFile;
	} 
	
	/**
	 * Get the Blake2b hash of default key file
	 * 
	 * @return	the hash of the key file
	 */
	private static byte[] getDefaultKeyFileHash() {
		String keyFileName = PeaProperties.getDefaultKeyFileName();
		if (keyFileName == null) {
			TestLog.ve(KeyFileDialogs.class, "Missing key file in properties " , 5);
			PeaDialog.showMessage(JREProperties.getMainWindow(), 
					PeaProperties.getVmBridge().translate("key_file") + " " + PeaProperties.getDefaultKeyFileName() + "\n"
					+ PeaProperties.getVmBridge().translate("file_not_found") +"\n\n"
					+ PeaProperties.getVmBridge().translate("choose_key_file"),
					PeaProperties.getVmBridge().translate("error"), 0);
			
			PeaFileChooser pfc = new PeaFileChooser(JREProperties.getMainWindow());
			pfc.setFileSelectionMode(PeaFileChooser.FILES_ONLY);
			pfc.customizeFileChooser(null, 
					PeaProperties.getVmBridge().translate("choose_key_file"),  
					true);
			pfc.showOpenDialog(JREProperties.getMainWindow());
			File keyFile = pfc.getSelectedFile();
			if (keyFile != null) {
				PeaProperties.setDefaultKeyFileName(keyFile.getAbsolutePath());
				return KeyFileHandler.readAndHashKeyFile(keyFile);
			} else {
				TestLog.v(KeyFileDialogs.class, "Opening key file cancelled...");
				return null;
			}
		}
		File keyFile = new File(keyFileName );
		if (keyFile.exists() == false) {
			TestLog.ve(KeyFileDialogs.class, "Key file does not exist: " + keyFileName, 5);
			PeaDialog.showMessage(JREProperties.getMainWindow(), 
					PeaProperties.getVmBridge().translate("key_file") + " " + PeaProperties.getDefaultKeyFileName() + "\n"
					+ PeaProperties.getVmBridge().translate("file_not_found") +"\n\n"
					+ PeaProperties.getVmBridge().translate("choose_key_file"),
					PeaProperties.getVmBridge().translate("error"), 0);
			
			PeaFileChooser pfc = new PeaFileChooser(JREProperties.getMainWindow());
			pfc.setFileSelectionMode(PeaFileChooser.FILES_ONLY);
			pfc.customizeFileChooser(null, 
					PeaProperties.getVmBridge().translate("choose_key_file"),  
					true);
			pfc.showOpenDialog(JREProperties.getMainWindow());
			keyFile = pfc.getSelectedFile();
			
			if (keyFile != null) {
				if (keyFile.canRead() == true) {
					probablyNewDefaultKeyFileName = keyFile.getAbsolutePath();
					return KeyFileHandler.readAndHashKeyFile(keyFile);
				} else {
					TestLog.v(KeyFileDialogs.class, "Opening key file cancelled...");
					return null;
				}
			} else {
				TestLog.v(KeyFileDialogs.class, "Opening key file cancelled...");
				return null;
			}
			
			//return getChosenKeyFileHash();
		}		
		if (keyFile.canRead() == false) {
			TestLog.ve(KeyFileDialogs.class, "Key file does not exist: " + keyFileName, 5);
			PeaDialog.showMessage(JREProperties.getMainWindow(), 
					PeaProperties.getVmBridge().translate("key_file") + " " + PeaProperties.getDefaultKeyFileName() + "\n"
					+ PeaProperties.getVmBridge().translate("no_access") ,
					PeaProperties.getVmBridge().translate("error"), 0);
			return null;
		}
		return KeyFileHandler.readAndHashKeyFile(keyFile);
	}
	
	/**
	 * Get the Blake2b hash of a manually chosen file
	 * 
	 * @return	the hash of the key file
	 */
	private static byte[] getChosenKeyFileHash() {
		PeaFileChooser pfc = new PeaFileChooser(JREProperties.getMainWindow());
		pfc.setFileSelectionMode(PeaFileChooser.FILES_ONLY);
		pfc.customizeFileChooser(null, 
				PeaProperties.getVmBridge().translate("choose_key_file"),  
				true);
		pfc.showOpenDialog(JREProperties.getMainWindow());
		File keyFile = pfc.getSelectedFile();
		if (keyFile != null) {
			return KeyFileHandler.readAndHashKeyFile(keyFile);
		} else {
			TestLog.v(KeyFileDialogs.class, "Opening key file cancelled...");
			return null;
		}
	}
	/**
	 * Get the probably new default key file name. 
	 * This value is set, when a default key file is missing
	 * and a new default key file was chosen. 
	 * 
	 * @return	the probably new key file name
	 */
	public static String getProbablyNewDefaultKeyFileName() {
		return KeyFileDialogs.probablyNewDefaultKeyFileName;
	}
	/**
	 * Reset the probably new default key file name. 
	 * This value is set, when a default key file is missing
	 * and a new default key file was chosen. 
	 */
	public static void resetProbablyNewDefaultKeyFileName() {
		KeyFileDialogs.probablyNewDefaultKeyFileName = null;
	}
}
