package cologne.eck.calendar_pea.control;

/*
 * PeaFactory - Production of Password Encryption Archives
 * Copyright (C) 2015  Axel von dem Bruch
 * 
 * This library is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published 
 * by the Free Software Foundation; either version 2 of the License, 
 * or (at your option) any later version.
 * This library is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * See:  http://www.gnu.org/licenses/gpl-2.0.html
 * You should have received a copy of the GNU General Public License 
 * along with this library.
 */

import java.nio.charset.Charset;
import java.time.DayOfWeek;
import java.time.format.DateTimeFormatter;
import java.time.format.FormatStyle;
import java.util.ArrayList;
import java.util.Arrays;



/*
 * This class contains variables and their Getters 
 * that can't be modified
 */
public class CalendarConstants {
	
	/*
	 * The default character set, used whenever a string
	 * is converted 
	 */
	private static final Charset UTF_8 = Charset.forName("UTF-8");	
	// Formatter for iCal date and time values
	private static final DateTimeFormatter iCalDateTimeFormatter = DateTimeFormatter.ofPattern("yyyyMMdd'T'HHmmss");
	// for all-day events: date only
	private static final DateTimeFormatter iCalDateFormatter = DateTimeFormatter.ofPattern("yyyyMMdd");
	
	
	// format for date and time:
	// used for one day events:
	private static final DateTimeFormatter fullDateFormatter = DateTimeFormatter.ofLocalizedDate(FormatStyle.FULL);
//	private static final DateTimeFormatter mediumDateFormatter = DateTimeFormatter.ofLocalizedDate(FormatStyle.MEDIUM);
	// used for several day events:
	private static final DateTimeFormatter shortDateFormatter = DateTimeFormatter.ofLocalizedDate(FormatStyle.SHORT);
	private static final DateTimeFormatter timeFormatter = DateTimeFormatter.ofPattern("HH:mm");

	// fur UTC format
//	private static final DateTimeFormatter UtcFormatter =
//			      DateTimeFormatter.ofPattern("yyyyMMdd'T'HHmmss'X'").withZone(ZoneId.of("UTC"));
	/* PRODID - must
	 * globally unique identifier, text that describes the product and version
	 */
	private static final String PROD_ID = "-//PEAFACTORY//CALENDAR LOCK PEA//EN";
	// the string to start a PCalendar
	//private static final String FIXED_CALENDAR_STRING = "BEGIN:VCALENDAR\r\nPRODID:" + PROD_ID + "\r\nVERSION:2.0\r\nCALSCALE:GREGORIAN\r\nMETHOD:PUBLISH\r\n";
	private static final String FIXED_CALENDAR_STRING = "BEGIN:VCALENDAR\r\nPRODID:" + PROD_ID + "\r\nVERSION:2.0\r\nCALSCALE:GREGORIAN\r\n";
	//private static final String FIXED_CALENDAR_STRING = "BEGIN:VCALENDAR\r\nCALSCALE:GREGORIAN\r\nVERSION:2.0\r\nPRODID:" + PROD_ID + "\r\nMETHOD:REQUEST\r\n";

	private static final String[] weekDaysICal = {"MO", "TU", "WE", "TH", "FR", "SA", "SU"};
	// this can't be final: initialize the first time it is used
	private static ArrayList<String> weekDaysList;
	
//	private final static String ECK_STRING = "@eck.cologne";
/*
 * rfc7986
 * This specification updates
   [RFC5545] by stating that "UID" values MUST NOT include any data that
   might identify a user, host, domain, or any other security- or
   privacy-sensitive information.  It is RECOMMENDED that calendar user
   agents now generate "UID" values that are hex-encoded random
   Universally Unique Identifier (UUID) values
 */
	
	
	/**
	 * Get the default charset: UTF-8
	 * 
	 * @return the UTF-8 charset
	 */
	public static Charset getCharset() {
		return UTF_8;
	}

	/**
	 * Get the formatter for iCal valid date and time values
	 * 
	 * @return	the formatter for date and time
	 */
	public static DateTimeFormatter getIcalDateTimeFormatter() {
		return iCalDateTimeFormatter;
	}

	/**
	 * Get the formatter for iCal valid date only values
	 * 
	 * @return	the formatter for date only
	 */
	public static DateTimeFormatter getIcalDateFormatter() {
		return iCalDateFormatter;
	}

	/**
	 * Get the fixed start String to write a PCalendar
	 * (contains scale, version,prodID, method)
	 * 
	 * @return	the fixed start String to write a Calendar
	 */
	public static String getFixedCalendarString() {
		return FIXED_CALENDAR_STRING;
	}
	/**
	 * Get the product ID for Calendar Lock PEA
	 * as defined in RFC 5545
	 * 
	 * @return	the product ID for Calendar Lock PEA
	 */
/*	public static final String getProdID() {
		return PROD_ID;
	} */

	public static DateTimeFormatter getFullDateFormatter() {
		return fullDateFormatter;
	}

//	public static DateTimeFormatter getMediumdateformatter() {
//		return mediumDateFormatter;
//	}
	/**
	 * Get the DateFormatter for short date without time
	 * 
	 * @return	the DateFormatter for short date without time
	 */
	public static DateTimeFormatter getShortDateFormatter() {
		return shortDateFormatter;
	}
	/**
	 * Get the formatter for dates with time
	 * 
	 * @return	 the formatter for dates with time
	 */
	public static DateTimeFormatter getTimeFormatter() {
		return timeFormatter;
	}
	/**
	 * Get the weekday string as used in iCalendar: 
	 * MO, TU, WE, TH, FR, SA, SU
	 * 
	 * @return	the weekday short names as used in iCalendar
	 */
	public static String[] getWeekdaysICal() {
		return weekDaysICal;
	}
	/**
	 * Get the short name of the weekday of a LocalDate
	 * as used in iCalendar: MO,TU.WE,TH,FR,SA,SU
	 * 
	 * @param date	the date to get the weekday
	 * 
	 * @return iCalendar weekday short name:
	 * 			MO,TU.WE,TH,FR,SA,SU
	 */
/*	public static String getDayOfWeekShortName(LocalDate date) {
		return weekDaysICal[date.getDayOfWeek().getValue() - 1];
	}*/
	/**
	 * Get the value of a weekDay, as used in iCalendar, 
	 *  MO is 1, not 0 !!!
	 * 
	 * @param weekDayShortName	the short name of the weekday:
	 * 							MO,TU.WE,TH,FR,SA,SU
	 * 
	 * @return	the value (not index) of the weekday: MO is 1
	 */
	public static int getDayOfWeekValue(String weekDayShortName) {
		if (weekDaysList == null) {
			weekDaysList = new ArrayList<>(Arrays.asList(weekDaysICal));
		}
		return weekDaysList.indexOf(weekDayShortName) + 1;// +1 because of index
	}
	
	/**
	 * Get the weekDay from a weekday short name as used in iCalendar
	 * 
	 * @param weekDayShortName	the short name of the weekday:
	 * 							MO,TU.WE,TH,FR,SA,SU
	 * 
	 * @return	the weekday as DayOfWeek
	 */
	public static DayOfWeek getDayOfWeek(String weekDayShortName) {
		if (weekDaysList == null) {
			weekDaysList = new ArrayList<>(Arrays.asList(weekDaysICal));
		}
		return DayOfWeek.of(weekDaysList.indexOf(weekDayShortName) + 1);
	}

	/**
	 * Get the string appendix for UID of 
	 * calendar and event: @eck.cologne
	 * 
	 * @return	the UID appendix for this application
	 */
	/*public static String getEckString() {
		return ECK_STRING;
	}*/
}
