package cologne.eck.calendar_pea.control;

/*
 * PeaFactory - Production of Password Encryption Archives
 * Copyright (C) 2015  Axel von dem Bruch
 * 
 * This library is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published 
 * by the Free Software Foundation; either version 2 of the License, 
 * or (at your option) any later version.
 * This library is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * See:  http://www.gnu.org/licenses/gpl-2.0.html
 * You should have received a copy of the GNU General Public License 
 * along with this library.
 */

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.net.MalformedURLException;
import java.time.LocalDate;
import java.time.format.DateTimeFormatter;
import java.util.Locale;
import java.util.ResourceBundle;

import cologne.eck.all_peas.vm_specific.JREProperties;
import cologne.eck.calendar_pea.gui.calendar_view.CalendarView;
import cologne.eck.calendar_pea.gui.calendar_view.WeekPanel;
import cologne.eck.tools.TestLog;


/*
 * This class contains variables and their Getters & Setters 
 * that might be changed by users
 */
public class CalendarSetting {
	

	/**
	 * The language bundle for Calendar Lock PEA
	 */
	private static ResourceBundle languagesBundle;		
	/**
	 * The last used view (day, week, month)
	 */
	private static CalendarView startView = new WeekPanel(LocalDate.now(), 
			new Dimension(JREProperties.getFrameDim()[0], JREProperties.getFrameDim()[1]));
	
	/**
	 * The default file name to store the encrypted content of all calendars 
	 */
	private static String defaultFolder;
	
	
	/**
	 * Formatter for Label of singleDayPanels in MonthPanel to show the date 
	 */
	//private static DateTimeFormatter dayDateFormatter = DateTimeFormatter.ofLocalizedDate(FormatStyle.LONG);
	// short names of Months
	private static DateTimeFormatter dayDateFormatter = DateTimeFormatter.ofPattern ( "dd MMM yyyy" ).withLocale ( Locale.getDefault() );

	/**
	 * Lines of days to show in the MonthPanel
	 */
	private static int numberOfLinesInMonthPanel = 5;
	
	/**
	 * the background color of days in MonthPanel
	 */
	private static Color dayBackgroundColor = new Color(252, 254, 253);//new Color(248, 248, 248);
	
	/**
	 * The background color for today in MonthPanel
	 */
	private static Color todayColor = new Color(255, 242, 204);
	
	/**
	 * The font used to show events in tool tips
	 */
	private static Font eventButtonFont = new Font(Font.SANS_SERIF, Font.PLAIN, 8);
	
	/**
	 * The time shown first in the scroll pane of DayView
	 */
	private static int viewStartHour = 7;
	
	
	//================== GETTERS & SETTERS ================//
	
	/**
	 * Set the language bundle for Calendar Lock PEA
	 * 
	 * @param newBundle	the language bundle to use
	 */
	protected static void setResourceBundle(ResourceBundle newBundle) {
		languagesBundle = newBundle;
	}
	/**
	 * Get the current language bundle. If no bundle
	 * was set before, it is set with default language
	 * 
	 * @return		the current language bundle
	 */
	public final static ResourceBundle getResourceBundle(){
		if (languagesBundle == null) { // set default language
			try {
				CalendarControl.setLanguagesBundle(null);
			} catch (MalformedURLException e) {
				TestLog.e(CalendarSetting.class, "Language support failed: "  + e.toString() + "\n  " + e.getMessage());
				e.printStackTrace();
			}
		}
		return languagesBundle;
	}
	/**
	 * Get the formatter to display a day date
	 * Default is LONG, possible formatters: FULL, MEDIUM, SHORT
	 * 
	 * @return 	the formatter to display a day date
	 */
	public static DateTimeFormatter getDayDateFormatter() {
		return dayDateFormatter;
	}
	/**
	 * Set the style of the formatter to display a day date. 
	 * Default is LONG, possible values are: FULL, MEDIUM, SHORT
	 * 
	 * @param 	the formatter style to display a day date
	 * 			Default is LONG, possible other values are: FULL, MEDIUM, SHORT
	 */
/*	public static void setDayDateFormatterStyle(String styleString) {
		if (styleString.toUpperCase().trim().equals("LONG")) {
			dayDateFormatter = DateTimeFormatter.ofLocalizedDate(FormatStyle.LONG);
		} else if (styleString.toUpperCase().trim().equals("FULL")) {
			dayDateFormatter = DateTimeFormatter.ofLocalizedDate(FormatStyle.FULL);
		} else if (styleString.toUpperCase().trim().equals("MEDIUM")) {
			dayDateFormatter = DateTimeFormatter.ofLocalizedDate(FormatStyle.MEDIUM);
		} else if (styleString.toUpperCase().trim().equals("SHORT")) {
			dayDateFormatter = DateTimeFormatter.ofLocalizedDate(FormatStyle.SHORT);
		} else {
			TestLog.e(getClass(), "Wrong style String to set format style of day date");
		}
	}*/
	/**
	 * Get the number of lines of days in MonthPanel.
	 * One line contains always seven days. 
	 * 
	 * @return	the number of lines to show
	 */
	public static int getNumberLinesInMonthPanel() {
		return numberOfLinesInMonthPanel;
	}
	/**
	 * Set the number of lines of days in MonthPanel.
	 * One line contains always seven days. 
	 * 
	 * @param	the number of lines to show
	 */
	public static void setNumberLinesInMonthPanel(int newNumberOfLinesInMonthPanel) {
		CalendarSetting.numberOfLinesInMonthPanel = newNumberOfLinesInMonthPanel;
	}

	/**
	 * Set the background color of days in MonthPanel
	 * 
	 * @return	the background color of days
	 */
	public static Color getDayBackgroundColor() {
		return dayBackgroundColor;
	}
	/**
	 * Set the background color of days in MonthPanel
	 * 
	 * @param	the background color of days
	 */
	public static void setDayBackgroundColor(Color dayBackgroundColor) {
		CalendarSetting.dayBackgroundColor = dayBackgroundColor;
	}
	/**
	 * Get the background color of the todays day panel in MonthPanel
	 * 
	 * @return	the color for today
	 */
	public static Color getTodayColor() {
		return todayColor;
	}
	/**
	 * Set the background color of the todays day panel in MonthPanel
	 * 
	 * @param	the color for today
	 */
	protected static void setTodayColor(Color todayColor) {
		CalendarSetting.todayColor = todayColor;
	}
	/**
	 * Get the font of event buttons in MonthPanel
	 * 
	 * @return	the font of event buttons
	 */
	public static Font getEventButtonFont() {
		return eventButtonFont;
	}
	/**
	 * Set the font of event buttons in MonthPanel
	 * 
	 * @param	the font of event buttons
	 */
	public static void setEventButtonFont(Font newEventButtonFont) {
		CalendarSetting.eventButtonFont = newEventButtonFont;
	}
	/**
	 * Get the default folder to save 
	 * the encrypted content of all calendars as files
	 * 
	 * @return the default folder to save calendars
	 */
	public static String getDefaultFolder() {
		return defaultFolder;
	}
	/**
	 * Set the default folder to save 
	 * the encrypted content of all calendars as files
	 * 
	 * @param the default folder to save calendars
	 */
	public static void setDefaultFolder(String _newDefaultFolder) {
		CalendarSetting.defaultFolder = _newDefaultFolder;
	}


	/**
	 * Get the first shown hour in scroll pane of DayView
	 * (position of scroll bar)
	 * 
	 * @return the first shown hour (position of scroll bar)
	 */
	public static int getViewStartHour() {
		return viewStartHour;
	}
	/**
	 * Set the first shown hour in scroll pane of DayView
	 * (position of scroll bar)
	 * 
	 * @param the first shown hour (position of scroll bar)
	 */
	public static void setViewStartHour(int viewStartHour) {
		CalendarSetting.viewStartHour = viewStartHour;
	}
	/**
	 * Get the used view (day, week, month)of the last session
	 * 
	 * @return	the last used view
	 */
	public static CalendarView getStartView() {
		return startView;
	}
	/**
	 * Set the used view (day, week, month)of the last session
	 * 
	 * @param startView	the last used view
	 */
	protected static void setStartView(CalendarView startView) {
		CalendarSetting.startView = startView;
	}
	/**
	 * Get the newest last modification of the currently used files
	 * when they were opened in this session
	 * 
	 * @return	the newest last modification of the currently 
	 * 			used files as long
	 */
/*	public static long getLastModAtStart() {
		return lastModAtStart;
	}*/
	/**
	 * Set the newest last modification of the currently used files
	 * when they were opened in this session
	 * 
	 * @param lastModAtStart	the newest last modification of the 
	 * 							currently used files as long
	 */
/*	protected static void setLastModAtStart(long lastModAtStart) {
		CalendarSetting.lastModAtStart = lastModAtStart;
	}*/
}
