package cologne.eck.calendar_pea.util;

/*
 * PeaFactory - Production of Password Encryption Archives
 * Copyright (C) 2015  Axel von dem Bruch
 * 
 * This library is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published 
 * by the Free Software Foundation; either version 2 of the License, 
 * or (at your option) any later version.
 * This library is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * See:  http://www.gnu.org/licenses/gpl-2.0.html
 * You should have received a copy of the GNU General Public License 
 * along with this library.
 */

import static org.junit.jupiter.api.Assertions.*;

import java.time.Instant;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.time.temporal.Temporal;

import org.junit.Assert;
import org.junit.jupiter.api.Test;

import cologne.eck.tools.TestLog;




class ICalDateTimeHelperTest {
	
	private static final String[] zoneIdStrings = { "America/New_York", "Europe/Zurich", 
			"US/Central", "Asia/Kuala_Lumpur", "Antarctica/Casey", "Australia/Adelaide", "Pacific/Auckland"
	};

	@Test
	final void testGetDatesFromList() {
		String[] listTests = {
				"EXDATE;VALUE=DATE-TIME;TZID=America/New_York:20140905T103000,20140906T103000",
				"EXDATE;VALUE=DATE-TIME;TZID=America/New_York;XXX=YYY:20140905T103000,20140906T103000",
				"RDATE;VALUE=DATE:20140905,20140906", // default Calendar Lock PEA
				"EXDATE:19960402T010000Z,19960403T010000Z,19960404T010000Z",// RFC5545 example
				"RDATE:19970714T123000Z",
				"RDATE;TZID=America/New_York:19970714T083000"
				//"RDATE;VALUE=PERIOD:19960403T020000Z/19960403T040000Z,19960404T010000Z/PT3H"
		};
		Temporal[][] listResults = {
				{LocalDate.parse("2014-09-05"), LocalDate.parse("2014-09-06")}, 
				{LocalDate.parse("2014-09-05"), LocalDate.parse("2014-09-06")}, 
				{LocalDate.parse("2014-09-05"), LocalDate.parse("2014-09-06")}, 
				{LocalDate.parse("1996-04-02"), LocalDate.parse("1996-04-03"), LocalDate.parse("1996-04-04")},
				{LocalDate.parse("1997-07-14")},
				{LocalDate.parse("1997-07-14")}
		};		
		for (int i=0; i< listTests.length;i++) {
			//TestLog.o(getClass(), "test " + i + ": " + listTests[i]);
			LocalDate[] dates = ICalDateTimeHelper.getDatesFromList(listTests[i]);
			assertNotNull(dates);
			for(int j=0;j<dates.length;j++) {
				Assert.assertEquals(dates[j], listResults[i][j]);
				//TestLog.o(getClass(), dates[j] + " = " + listResults[i][j]);
			}
		}
		TestLog.o(getClass(), "getDatesFromList() ok");
	}

	@Test
	final void testGetDateTime() {

		String[] testStrings = {
				"19960401T150000Z",
				";VALUE=DATE:19980704",
				";TZID=Europe/Berlin:20181227",
				"20160302T000000",
				";VALUE=DATE-TIME;TZID=America/New_York:20140905T103000",
				";VALUE=DATE-TIME;TZID=America/New_York;XXX=YYY:20140905T103000"			
		};
		Temporal[] results = {
				LocalDateTime.parse("1996-04-01T17:00"), 
				LocalDate.parse("1998-07-04"),
				LocalDate.parse("2018-12-27"),
				LocalDateTime.parse("2016-03-02T00:00"),
				LocalDateTime.parse("2014-09-05T16:30"),
				LocalDateTime.parse("2014-09-05T16:30")
		};
		for (int i=0;i<testStrings.length;i++) {
			Assert.assertEquals(ICalDateTimeHelper.getDateTime(testStrings[i]), results[i]);
		}	
		TestLog.o(getClass(), "getDateTime() ok");
	}

	@Test
	final void testZonedDateTimeToLocalDateTime() {

		// default time zone:
		ZonedDateTime zdtDefault = ZonedDateTime.now();
		//TestLog.o(getClass(), ICalDateTimeHelper.getDefaultZoneId());

		for (int i = 0; i < zoneIdStrings.length; i++) {
			ZonedDateTime zdt = zdtDefault.withZoneSameInstant(ZoneId.of(zoneIdStrings[i])); 
			//TestLog.o(getClass(), zdt);
			Assert.assertEquals(zdtDefault.toLocalDateTime(), ICalDateTimeHelper.zonedDateTimeToLocalDateTime(zdt));
			Assert.assertEquals(LocalDateTime.now().withSecond(0).withNano(0), ICalDateTimeHelper.zonedDateTimeToLocalDateTime(zdt).withSecond(0).withNano(0));
		}
		TestLog.o(getClass(), "zonedDateTimeToLocalDateTime() ok");
	}

	@Test
	final void testZonedDateTimeToLocalDate() {

		// default time zone:

		ZonedDateTime zdtDefault = ZonedDateTime.now();
		//TestLog.o(getClass(), ICalDateTimeHelper.getDefaultZoneId());

		for (int i = 0; i < zoneIdStrings.length; i++) {
			ZonedDateTime zdt = zdtDefault.withZoneSameInstant(ZoneId.of(zoneIdStrings[i])); 
			//TestLog.o(getClass(), zdt);
			Assert.assertEquals(zdtDefault.toLocalDate(), ICalDateTimeHelper.zonedDateTimeToLocalDate(zdt));
			Assert.assertEquals(LocalDate.now(), ICalDateTimeHelper.zonedDateTimeToLocalDate(zdt));
		}
		TestLog.o(getClass(), "zonedDateTimeToLocalDate() ok");
	}

	@Test
	final void testDurationToDateEnd() {

		String[][] durationTests = {
				{"DTSTART;VALUE=DATE:20181101", "P1W"}, 
				{"DTSTART;VALUE=DATE:20181102", "P2D"},
				{"DTSTART;VALUE=DATE:20181105", "P1W2D"},
				{"DTSTART:20181114T120000", "PT1H"},
				{"DTSTART:20181116T121000", "PT11H5M0S"},
				{"DTSTART:20181118T114500", "P5DT5H0M20S"},
				{"DTSTART:20181202T121000", "PT100H100M0S"}
		};
		String[] durationResults = {"2018-11-08T00:00", "2018-11-04T00:00", "2018-11-14T00:00",  
				"2018-11-14T13:00", "2018-11-16T23:15", "2018-11-23T16:45:20", "2018-12-06T17:50" };
		for (int i = 0; i < durationTests.length; i++) {
			Temporal tmp = ICalDateTimeHelper.getDateTime(durationTests[i][0]);
			if (tmp instanceof LocalDate) {
				tmp = ((LocalDate) tmp).atStartOfDay();
			}
			assertEquals(LocalDateTime.parse(durationResults[i]) , ICalDateTimeHelper.durationToDateEnd((LocalDateTime)tmp, durationTests[i][1]));
			//TestLog.o(getClass(), "   " + ICalDateTimeHelper.durationToDateEnd((LocalDateTime)tmp, durationTests[i][1]));
		}
		TestLog.o(getClass(), "durationToDateEnd()  ok");
	}

	@Test
	final void testGetInstantFromUtcString() {
		String test = "19970714T123000Z";
		//TestLog.o(getClass(), "--- " + ICalDateTimeHelper.getInstantFromUtcString(test));
		//Instant.parse("1997-07-14T12:30:00Z");
		assertEquals(Instant.parse("1997-07-14T12:30:00Z"), ICalDateTimeHelper.getInstantFromUtcString(test));
		TestLog.o(getClass(), "getInstantFromUtcString() ok");
	}

	@Test
	final void testGetZoneIdFromString() {
		for (int i = 0; i < zoneIdStrings.length; i++) {
			ZoneId zid = ICalDateTimeHelper.getZoneIdFromString(zoneIdStrings[i], null, null);
			assertNotNull(zid);
		}
		TestLog.o(getClass(), "getZoneIdFromString() ok");
	}
}
