package cologne.eck.peafactory.crypto;

/*
 * Peafactory - Production of Password Encryption Archives
 * Copyright (C) 2020  Axel von dem Bruch
 * 
 * This library is free software; you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published 
 * by the Free Software Foundation; either version 2 of the License, 
 * or (at your option) any later version.
 * This library is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 * See:  http://www.gnu.org/licenses/gpl-2.0.html
 * You should have received a copy of the GNU General Public License 
 * along with this library.
 */

import java.nio.charset.Charset;

import org.bouncycastle.crypto.BlockCipher;
import org.bouncycastle.crypto.Digest;
import org.bouncycastle.crypto.digests.Blake2bDigest;
import org.bouncycastle.crypto.digests.SHA3Digest;
import org.bouncycastle.crypto.engines.AESEngine;
import org.bouncycastle.crypto.engines.ThreefishEngine;

import cologne.eck.all_peas.bridge.VMBridge;
import cologne.eck.all_peas.data.PeaProperties;
import cologne.eck.peafactory.crypto.kdf.Argon2KDF;
import cologne.eck.peafactory.crypto.kdf.CatenaKDF;
import cologne.eck.peafactory.crypto.kdf.KeyDerivation;
import cologne.eck.tools.ExceptionHelper;
import cologne.eck.tools.TestLog;
import cologne.eck.tools.UnexpectedValueException;


public class AlgoParamHandler {

	private final static int ALGO_PARAM_STRING_SIZE = 24;
	
	private final static String V1_ALGOPARAM_STRING = "CDUB2TE012$2$-----------";
	
	private final static String[] KDFS = {
			"A-d", "Ai-", "Aid",// Argon 2
			"CDU", "CDA", "CBU", "CBA", // Catena Dragonfly/Butterfly fUll/fAst
			"SCR", "Ly2" };
	private final static String[] HASHS = { "B2", "S3", "S2"};// Blake2b, SHA-3, SHA-512
	//private final static String[] CIPHERS = { "T", "A", "S"}; // Threefish, AES, Serpent
	//private final static String[] MODES = { "e", "G", "N"}; // EAX, GCM, Null (CAESAR competition algos)
	//private final static String[] KEY_FILE_FLAG = {"0", "1", "2"}; // no key file, default key file, user selected key file
	
	/**
	 * Read an algo-parameter String, set the algorithms and parameters
	 * 
	 * @param algoParams	the algo-parameter String
	 * @param setDefault	true: set only default properties, 
	 * 						false: set for next usage
	 * 
	 * @return	null for success, an error message otherwise
	 */
	public static String readString(String algoParams, boolean setDefaults) {
		
		if (algoParams.length() != ALGO_PARAM_STRING_SIZE) {
			new UnexpectedValueException("String", "algoParams", "invalidLength: " + algoParams + ", len: " + algoParams.length()).printDescription();
			return "Invalid length: " + algoParams.length();
		}
		TestLog.v(AlgoParamHandler.class, algoParams, 2);
		
		String paramString = algoParams.substring(8, algoParams.length());
		//TestLog.o(getClass(), "paramString: " + paramString);
		String[] params = paramString.split("\\$");
		//for(String s : params)TestLog.o(getClass(), s);
		if (params.length < 3) { // at least memory and time cost parameter (for Bcrypt memory constant)
			new UnexpectedValueException("String[]", "params", "invalid size: " + params.length).printDescription();
			return "Invalid parameters: " + paramString;
		}
		int tCost = 0,  mCost = 0, pCost = 0;
		try {
			mCost = Integer.parseInt(params[0], 16);
			tCost = Integer.parseInt(params[1], 16);
			//KeyDerivation.getKdf().setmCost(Integer.parseInt(params[0], 16));
			//KeyDerivation.getKdf().settCost(Integer.parseInt(params[1], 16));
			// the last part is always trailing "---"
			int paramNumber = params.length;
			if (paramNumber > 3) {
				int p = Integer.parseInt(params[2], 16);
				pCost = p;
				//KeyDerivation.getKdf().setArg3(p);
			} 

		} catch (NumberFormatException e){
			TestLog.e(AlgoParamHandler.class, e.toString() + " - " + e.getLocalizedMessage(), 5);
			return "Invalid parameters";
		}
		
		String kdf = algoParams.substring(0, 3);
		if (kdf.equals(KDFS[0])) {
			if (setDefaults == true) {
				PeaProperties.setDefaultKDFInstanceAppendix("d");
				PeaProperties.setDefaultMCost(mCost);
				PeaProperties.setDefaultTCost(tCost);
				PeaProperties.setDefaultParallel(pCost);
				PeaProperties.setDefaultKDF(new Argon2KDF());
			} else {
				Argon2KDF.setType(0);
				Argon2KDF.setMemoryCost(mCost);
				Argon2KDF.setIterations(tCost);
				Argon2KDF.setLanes(pCost);
				KeyDerivation.setKdf(new Argon2KDF());
			}
		} else if (kdf.equals(KDFS[1])) {
			if (setDefaults == true) {
				PeaProperties.setDefaultKDFInstanceAppendix("i");
				PeaProperties.setDefaultMCost(mCost);
				PeaProperties.setDefaultTCost(tCost);
				PeaProperties.setDefaultParallel(pCost);
				PeaProperties.setDefaultKDF(new Argon2KDF());
			} else {
				Argon2KDF.setType(1);
				Argon2KDF.setMemoryCost(mCost);
				Argon2KDF.setIterations(tCost);
				Argon2KDF.setLanes(pCost);
				KeyDerivation.setKdf(new Argon2KDF());
			}
		} else if (kdf.equals(KDFS[2])) {
			if (setDefaults == true) {
				PeaProperties.setDefaultKDFInstanceAppendix("id");
				PeaProperties.setDefaultMCost(mCost);
				PeaProperties.setDefaultTCost(tCost);
				PeaProperties.setDefaultParallel(pCost);
				PeaProperties.setDefaultKDF(new Argon2KDF());
			} else {
				Argon2KDF.setType(2);
				Argon2KDF.setMemoryCost(mCost);
				Argon2KDF.setIterations(tCost);
				Argon2KDF.setLanes(pCost);
				KeyDerivation.setKdf(new Argon2KDF());
			}
		} else if (kdf.equals(KDFS[3])) {
			if (setDefaults == true) {
				PeaProperties.setDefaultKDFInstanceAppendix("Dragonfly-Full");
				PeaProperties.setDefaultMCost(mCost);
				PeaProperties.setDefaultTCost(tCost);
				PeaProperties.setDefaultParallel(0);
				PeaProperties.setDefaultKDF(new CatenaKDF());
			} else {
				CatenaKDF.setVersionID("Dragonfly-Full");
				CatenaKDF.setGarlicDragonfly(mCost);
				CatenaKDF.setLambdaDragonfly(tCost);
				KeyDerivation.setKdf( new CatenaKDF() );
			}
		} else if (kdf.equals(KDFS[4])) {
			if (setDefaults == true) {
				PeaProperties.setDefaultKDFInstanceAppendix("Dragonfly");
				PeaProperties.setDefaultMCost(mCost);
				PeaProperties.setDefaultTCost(tCost);
				PeaProperties.setDefaultParallel(0);
				PeaProperties.setDefaultKDF(new CatenaKDF());
			} else {
				CatenaKDF.setVersionID("Dragonfly");
				CatenaKDF.setGarlicDragonfly(mCost);
				CatenaKDF.setLambdaDragonfly(tCost);				
				KeyDerivation.setKdf( new CatenaKDF() );
			}
		} else if (kdf.equals(KDFS[5])) {
			if (setDefaults == true) {
				PeaProperties.setDefaultKDFInstanceAppendix("Butterfly-Full");
				PeaProperties.setDefaultMCost(mCost);
				PeaProperties.setDefaultTCost(tCost);
				PeaProperties.setDefaultParallel(0);
				PeaProperties.setDefaultKDF(new CatenaKDF());
			} else {
				CatenaKDF.setVersionID("Butterfly-Full");
				CatenaKDF.setGarlicButterfly(mCost);
				CatenaKDF.setLambdaButterfly(tCost);
				KeyDerivation.setKdf( new CatenaKDF() );
			}
		} else if (kdf.equals(KDFS[6])){
			if (setDefaults == true) {
				PeaProperties.setDefaultKDFInstanceAppendix("Butterfly");
				PeaProperties.setDefaultMCost(mCost);
				PeaProperties.setDefaultTCost(tCost);
				PeaProperties.setDefaultParallel(0);
				PeaProperties.setDefaultKDF(new CatenaKDF());
			} else {
				CatenaKDF.setVersionID("Butterfly");
				CatenaKDF.setGarlicButterfly(mCost);
				CatenaKDF.setLambdaButterfly(tCost);
				KeyDerivation.setKdf( new CatenaKDF() );
			}
		} else {
			new UnexpectedValueException("String", "kdf", "does not match algos: " + kdf).printDescription();
			return "Invalid key derivation function: " + kdf;
		}		
		String hash = algoParams.substring(3, 5);
		if (hash.equals(HASHS[0])) {
			if (setDefaults == true) {
				PeaProperties.setDefaultDigest(new Blake2bDigest());
			} else {
				HashStuff.setHashAlgo(new Blake2bDigest());
				//HashStuff.setHMacAlgo(new Blake2bDigest());
			}
		} else if (hash.equals(HASHS[1])) {
			if (setDefaults == true) {
				PeaProperties.setDefaultDigest(new SHA3Digest());
			} else {
				HashStuff.setHashAlgo(new SHA3Digest());
				//HashStuff.setHMacAlgo(new SHA3Digest());
			}
		} else {
			new UnexpectedValueException("String", "hash", "does not match algos: " + hash).printDescription();
			return "Invalid hash function: " + hash;
		}
		char cipher = algoParams.charAt(5);
		if (cipher == 'T') {
			if (setDefaults == true) {
				PeaProperties.setDefaultCipher(new ThreefishEngine(512));
			} else {
				CipherStuff.setCipherAlgo(new ThreefishEngine(512));
			}
		} else if (cipher == 'A') {
			if (setDefaults == true) {
				PeaProperties.setDefaultCipher(new AESEngine());
			} else {
				CipherStuff.setCipherAlgo(new AESEngine());
			}
		} else {
			new UnexpectedValueException("char", "cipher", "does not match algos: " + cipher).printDescription();
			return "Invalid cipher: " + cipher;
		}
		char mode = algoParams.charAt(6);
		if (mode == 'E') {
			if (setDefaults == true) {
				PeaProperties.setDefaultMode(new EAXMode());
			} else {
				CipherStuff.setCipherMode(new EAXMode());
			}
		} else {
			new UnexpectedValueException("char", "mode", "does not match algos: " + mode).printDescription();
			return "Invalid mode: " + mode;
		}
		char keyFileFlag = algoParams.charAt(7);
		if (! Character.isDigit(keyFileFlag)) {
			new UnexpectedValueException("char", "keyFileFlag", "is not digit: " + keyFileFlag
					+ " algo parameter string: " + algoParams).printDescription();
			return "Invalid key file flag: " + keyFileFlag;
		}
		int kff = Character.getNumericValue(keyFileFlag);
		if (kff >= 0 && kff <= 2) {//== '0' || kff == '1' || kff == '2' ) {
			if (setDefaults == true) {
				PeaProperties.setDefaultKeyFileFlag(kff);
			} else {
				PeaProperties.setKeyFileFlag(kff);
			}
		} else {
			TestLog.o(AlgoParamHandler.class, "kff: " + (kff < 0) + ", " + (kff > 2));
			new UnexpectedValueException("int", "kff", "is invalid: " + kff
					+ " algo parameter string: " + algoParams).printDescription();
			return "Invalid key file flag: " + keyFileFlag;
		}
		return null;
	}
	/**
	 * Get the cipher algorithm from an algorithm-parameter string
	 * 
	 * @param algoParams	the string as ASCII encoded string
	 * 
	 * @return	the cipher algorithm
	 */
	public static BlockCipher getCipherFromString(String algoParams) {
		if (algoParams.length() != ALGO_PARAM_STRING_SIZE) {
			new UnexpectedValueException("String", "algoParams", "invalidLength: " + algoParams + ", len: " + algoParams.length()).printDescription();
			TestLog.e(AlgoParamHandler.class, "Invalid length: " + algoParams.length());
			return null;
		}
		char cipher = algoParams.charAt(5);
		if (cipher == 'T') {
			return new ThreefishEngine(512);
		} else if (cipher == 'A') {
			return new AESEngine();
		} else {
			new UnexpectedValueException("char", "cipher", "does not match algos: " + cipher).printDescription();
			return null;
		}
	}
	/**
	 * Get the cipher algorithm from an algorithm-parameter string
	 * 
	 * @param algoParams	the string as ASCII encoded string
	 * 
	 * @return	the cipher algorithm
	 */
	public static Digest getHashFromString(String algoParams) {
		if (algoParams.length() != ALGO_PARAM_STRING_SIZE) {
			new UnexpectedValueException("String", "algoParams", "invalidLength: " + algoParams + ", len: " + algoParams.length()).printDescription();
			TestLog.e(AlgoParamHandler.class, "Invalid length: " + algoParams.length());
			return null;
		}
		String hash = algoParams.substring(3, 5);
		if (hash.equals(HASHS[0])) {
			return new Blake2bDigest();
		} else if (hash.equals(HASHS[1])) {
			return new SHA3Digest();
		} else {
			new UnexpectedValueException("String", "hash", "does not match algos: " + hash).printDescription();
			return null;
		}
	}
	
	/**
	 * Get the String for current algorithms and parameters
	 * 
	 * @param fromDefaults	true: get values from defaults in PeaProperties, 
	 * 						false: get current algorithms and parameters
	 * 
	 * @return	the String for current algorithms and parameters
	 */
	public static String writeString(boolean fromDefaults) {
		String kdfFlag = null, hash = null, cipher = null, mode = null, keyFile = null;
		int m = 0, t = 0, p = 0;
		Digest digest = null;
		if (fromDefaults == true) {
			digest = PeaProperties.getDefaultDigest();
		} else {
			digest = HashStuff.getHashAlgo();
		}
		if (digest instanceof Blake2bDigest) {
			hash = "B2";
		} else if (digest instanceof SHA3Digest) {
			hash = "S3";
		} else {
			new UnexpectedValueException("Digest", "digest", "is invalid: " + digest.getAlgorithmName()).printDescription();
			return null;
		}
		BlockCipher cipherAlgo = null;
		if (fromDefaults == true) {
			cipherAlgo = PeaProperties.getDefaultCipher();
		} else {
			cipherAlgo = CipherStuff.getCipherAlgo();
		}
		if (cipherAlgo instanceof ThreefishEngine) {
			cipher = "T";
		} else if (cipherAlgo instanceof AESEngine) {
			cipher = "A";
		} else {
			new UnexpectedValueException("BlockCipher", "cipherAlgo", "is invalid: " + cipherAlgo.getAlgorithmName()).printDescription();
			return null;
		}
		AuthenticatedEncryption ae = null;
		if (fromDefaults == true) {
			ae = PeaProperties.getDefaultMode();
		} else {
			ae = CipherStuff.getCipherMode();
		}
		if (ae instanceof EAXMode) {
			mode = "E";
		} else {
			new UnexpectedValueException("AuthenticatedEncryption", "ae", "is invalid: " + ae.getClass().getSimpleName()).printDescription();
			return null;
		}

		if (fromDefaults == true) {
			keyFile = "" + PeaProperties.getDefaultKeyFileFlag();
		} else {
			if (PeaProperties.getKeyFileFlag() > 2) {
				new UnexpectedValueException("int", "keyFileFlag", "is invalid: " + PeaProperties.getKeyFileFlag() + " -" + ae.getClass().getSimpleName()).printDescription();
				return null;
			}
			keyFile = "" + PeaProperties.getKeyFileFlag();
		}

		KeyDerivation kdf = null;
		if (fromDefaults == true) {
			kdf = PeaProperties.getDefaultKDF();
			if (kdf == null) {
				new UnexpectedValueException("KeyDerivation", "kdf", "is null.").printDescription();
				return null;
			}

		} else {
			kdf = KeyDerivation.getKdf();
		}
		if (kdf instanceof Argon2KDF) {
			if (fromDefaults == true) {
				String defaultInstance = PeaProperties.getDefaultKDFInstanceAppendix();
				if (defaultInstance.equals("i")) {
					kdfFlag = "Ai-";
				} else if (defaultInstance.equals("d")) {
					kdfFlag = "A-d";
				} else if (defaultInstance.equals("id")) {
					kdfFlag = "Aid";
				} else {
					new UnexpectedValueException("String", "kdf appendix", "is null.").printDescription();
					return null;
				}
				m = PeaProperties.getDefaultMCost();
				t = PeaProperties.getDefaultTCost();
				p = PeaProperties.getDefaultParallel();
			} else {
				int type = Argon2KDF.getType();
				if (type == 0) {
					kdfFlag = "A-d";
				} else if (type == 1) {
					kdfFlag = "Ai-";
				} else if (type == 2) {
					kdfFlag = "Aid";
				}
				m = Argon2KDF.getMemoryCost();
				t = Argon2KDF.getIterations();
				p = Argon2KDF.getLanes();
			}
		} else if (kdf instanceof CatenaKDF) {
			if (fromDefaults == true) {
				String defaultInstance = PeaProperties.getDefaultKDFInstanceAppendix();
				if (defaultInstance.equals("Dragonfly")) {
					kdfFlag = "CDA";
				} else if (defaultInstance.equals("Dragonfly-Full")) {
					kdfFlag = "CDU";
				} else if (defaultInstance.equals("Butterfly")) {
					kdfFlag = "CBA";
				} else if (defaultInstance.equals("Butterfly-Full")) {
					kdfFlag = "CBU";
				} else {
					new UnexpectedValueException("String", "versionID", "invalid Catena version: " + defaultInstance).printDescription();
					return null;
				}
				m = PeaProperties.getDefaultMCost();
				t = PeaProperties.getDefaultTCost();
				p = 0;
			} else {
				String versionID = CatenaKDF.getVersionID();
				//TestLog.o(getClass(), "versionID: " + versionID);
				p = KeyDerivation.getKdf().getArg3();// 0
				if (versionID.startsWith("Dragonfly")) {
					m = CatenaKDF.getGarlicDragonfly();
					t = CatenaKDF.getLambdaDragonfly();				
					if (versionID.equals("Dragonfly-Full")) {
						kdfFlag = "CDU";
					} else if (versionID.equals("Dragonfly")) {
						kdfFlag = "CDA";
					} else {
						new UnexpectedValueException("String", "versionID", "invalid Catena version: " + versionID).printDescription();
						return null;
					}
				} else if (versionID.startsWith("Butterfly")) {
					m = CatenaKDF.getGarlicButterfly();
					t = CatenaKDF.getLambdaButterfly();
					if (versionID.equals("Butterfly-Full")) {
						kdfFlag = "CBU";
					} else if (versionID.equals("Butterfly")) {
						kdfFlag = "CBA";
					} else {
						new UnexpectedValueException("String", "versionID", "invalid Catena version: " + versionID).printDescription();
						return null;
					}
				} else {
					new UnexpectedValueException("String", "version", "invalid Catena version: " + versionID).printDescription();
					return null;
				}
			}
		} else {
			new UnexpectedValueException("KeyDerivation", "kdf", "is invalid: " + kdf).printDescription();
			return null;
		}
		StringBuilder sb = new StringBuilder();
		//kdfFlag = null, hash = null, cipher = null, mode = null, keyFile = null;
		sb.append(kdfFlag);
		sb.append(hash);
		sb.append(cipher);
		sb.append(mode);
		sb.append(String.valueOf(keyFile));
		sb.append(Integer.toHexString(m) + "$");
		//TestLog.o(getClass(), "m: " + m + ", hex: " + Integer.toHexString(m));
		sb.append(Integer.toHexString(t) + "$");
		//TestLog.o(getClass(), "t: " + t + ", hex: " + Integer.toHexString(t));
		//TestLog.o(getClass(), "p: " +p+", " + new String(sb));
		if (p != 0) {			
			sb.append(Integer.toHexString(p) + "$");
		}
		while (sb.length() < ALGO_PARAM_STRING_SIZE) {
			sb.append('-');
		}
		String algoParamString = new String(sb);
		TestLog.v(AlgoParamHandler.class, algoParamString, 2);
		//TestLog.o(getClass(), algoParamString);
		return algoParamString;
	}
	/**
	 * Check if a String is a valid algorithm-parameter string. 
	 * 
	 * @param algoParams	the string to check
	 * 
	 * @return	null for success, an error message otherwise
	 */
	public static String checkString(String algoParams) {
		
		if (algoParams.length() != ALGO_PARAM_STRING_SIZE) {
			new UnexpectedValueException("String", "algoParams", "invalidLength: " + algoParams + ", len: " + algoParams.length()).printDescription();
			return "Invalid length: " + algoParams.length();
		}
	
		String paramString = algoParams.substring(8, algoParams.length());
		//TestLog.o(getClass(), "paramString: " + paramString);
		String[] params = paramString.split("\\$");
		//for(String s : params)TestLog.o(getClass(), s);
		if (params.length < 3) { // at least memory and time cost parameter (for Bcrypt memory constant)
			TestLog.e(AlgoParamHandler.class, "Unexpected error");
			TestLog.ve(AlgoParamHandler.class, "algorithm parameter string failed: " + paramString
					+ " from String " + algoParams + ", split in " + params.length + " parts", 5);
			new UnexpectedValueException("String[]", "params", "invalid size: " + params.length).printDescription();
			return "Invalid parameters: " + paramString;
		}
		int tCost = 0,  mCost = 0, pCost = 0;
		try {
			mCost = Integer.parseInt(params[0], 16);
			tCost = Integer.parseInt(params[1], 16);
			// the last part is always trailing "---"
			int paramNumber = params.length;
			if (paramNumber > 3) {
				int p = Integer.parseInt(params[2], 16);
				pCost = p;
				//KeyDerivation.getKdf().setArg3(p);
			} 
		} catch (NumberFormatException e){
			TestLog.e(AlgoParamHandler.class, new ExceptionHelper(e).getInfos());
			TestLog.e(AlgoParamHandler.class, tCost + ", " + mCost + ", " + pCost);
			return "Invalid parameters";
		}
		
		String kdf = algoParams.substring(0, 3);
		if (kdf.equals(KDFS[0])) {
		} else if (kdf.equals(KDFS[1])) {
		} else if (kdf.equals(KDFS[2])) {
		} else if (kdf.equals(KDFS[3])) {
		} else if (kdf.equals(KDFS[4])) {
		} else if (kdf.equals(KDFS[5])) {
		} else if (kdf.equals(KDFS[6])){
		} else {
			new UnexpectedValueException("String", "kdf", "does not match algos: " + kdf).printDescription();
			return "Invalid key derivation function: " + kdf;
		}		
		String hash = algoParams.substring(3, 5);
		if (hash.equals(HASHS[0])) {
		} else if (hash.equals(HASHS[1])) {
		} else {
			new UnexpectedValueException("String", "hash", "does not match algos: " + hash).printDescription();
			return "Invalid hash function: " + hash;
		}
		char cipher = algoParams.charAt(5);
		if (cipher == 'T') {
		} else if (cipher == 'A') {
		} else {
			new UnexpectedValueException("char", "cipher", "does not match algos: " + cipher).printDescription();
			return "Invalid cipher: " + cipher;
		}
		char mode = algoParams.charAt(6);
		if (mode == 'E') {
		} else {
			new UnexpectedValueException("char", "mode", "does not match algos: " + mode).printDescription();
			return "Invalid mode: " + mode;
		}
		char keyFileFlag = algoParams.charAt(7);
		if (! Character.isDigit(keyFileFlag)) {
			new UnexpectedValueException("char", "keyFileFlag", "is not digit: " + keyFileFlag
					+ " algo parameter string: " + algoParams).printDescription();
			return "Invalid key file flag: " + keyFileFlag;
		}
		int kff = Character.getNumericValue(keyFileFlag);
		if (kff >= 0 && kff <= 2) {//== '0' || kff == '1' || kff == '2' ) {
		} else {
			TestLog.o(AlgoParamHandler.class, "kff: " + (kff < 0) + ", " + (kff > 2));
			new UnexpectedValueException("int", "kff", "is invalid: " + kff
					+ " algo parameter string: " + algoParams).printDescription();
			return "Invalid key file flag: " + keyFileFlag;
		}
		return null;
	}
	
	public static final String getReadableString(String algoParams, boolean multiLine) {
		//StringBuilder sb = new StringBuilder();
		String keyDerivation = "";
		String mem = "";
		String time = "";
		String parallel = "";
		String cipherAlgo = "";
		String cipherMode = "";
		String hashAlgo = "";
		String keyFile = "";
		
		if (algoParams.length() != ALGO_PARAM_STRING_SIZE) {
			new UnexpectedValueException("String", "algoParams", "invalidLength: " + algoParams + ", len: " + algoParams.length()).printDescription();
			return "Invalid length: " + algoParams.length();
		}
		String kdf = algoParams.substring(0, 3);
		if (kdf.equals(KDFS[0])) {
			keyDerivation = "Argon2d";
		} else if (kdf.equals(KDFS[1])) {
			keyDerivation = "Argon2i";
		} else if (kdf.equals(KDFS[2])) {
			keyDerivation = "Argon2id";
		} else if (kdf.equals(KDFS[3])) {
			keyDerivation = "Dragonfly-Full";
		} else if (kdf.equals(KDFS[4])) {
			keyDerivation = "Dragonfly-Fast";
		} else if (kdf.equals(KDFS[5])) {
			keyDerivation = "Butterfly-Full";
		} else if (kdf.equals(KDFS[6])){
			keyDerivation = "Butterfly-Fast";
		} else {
			new UnexpectedValueException("String", "kdf", "does not match algos: " + kdf).printDescription();
			return "Invalid key derivation function: " + kdf;
		}		
		String hash = algoParams.substring(3, 5);
		if (hash.equals(HASHS[0])) {
			hashAlgo = "Blake2b";
		} else if (hash.equals(HASHS[1])) {
			hashAlgo = "SHA-3";
		} else {
			new UnexpectedValueException("String", "hash", "does not match algos: " + hash).printDescription();
			return "Invalid hash function: " + hash;
		}
		char cipher = algoParams.charAt(5);
		if (cipher == 'T') {
			cipherAlgo = "Threefish-512";
		} else if (cipher == 'A') {
			cipherAlgo = "AES-256";
		} else {
			new UnexpectedValueException("char", "cipher", "does not match algos: " + cipher).printDescription();
			return "Invalid cipher: " + cipher;
		}
		char mode = algoParams.charAt(6);
		if (mode == 'E') {
			cipherMode = "EAX mode";
		} else {
			new UnexpectedValueException("char", "mode", "does not match algos: " + mode).printDescription();
			return "Invalid mode: " + mode;
		}
		char keyFileFlag = algoParams.charAt(7);
		if (keyFileFlag == '0' ) {
			keyFile = "no key file";
		} else if (keyFileFlag == '1' ) {
			keyFile = "default key file";
		} else if (keyFileFlag == '2' ) {
			keyFile = "chosen key file";
		} else {
			new UnexpectedValueException("char", "keyFileFlag", "is invalid: " + keyFileFlag
					+ "\n" + algoParams).printDescription();
			return "Invalid key file flag: " + keyFileFlag;
		}
		
		String paramString = algoParams.substring(8, algoParams.length());
		//TestLog.o(getClass(), "paramString: " + paramString);
		String[] params = paramString.split("\\$");
		//for(String s : params)TestLog.o(getClass(), s);
		if (params.length < 3) { // at least memory and time cost parameter (for Bcrypt memory constant)
			new UnexpectedValueException("String[]", "params", "invalid size: " + params.length).printDescription();
			return "Invalid parameters: " + paramString;
		}
		try {
			mem = "" + Integer.parseInt(params[0], 16);
			time = "" + Integer.parseInt(params[1], 16);
			//sb.append("memory " + Integer.parseInt(params[0], 16) + ", ");
			//sb.append("time " + Integer.parseInt(params[1], 16) );
			// the last part is always trailing "---"
			if (params.length > 3) {
				//sb.append(", parallel " + Integer.parseInt(params[2], 16));
				parallel = "" + Integer.parseInt(params[2], 16);
			} /*else if (params.length > 4) {
				int y = Integer.parseInt(params[3], 16);
				KeyDerivation.getKdf().setArg4(y);
			} else if (params.length > 5) {
				int x = Integer.parseInt(params[4], 16);
				KeyDerivation.getKdf().setArg5(x);
			} else if (params.length > 6) {
				int x = Integer.parseInt(params[5], 16);
				KeyDerivation.getKdf().setArg6(x);
			} else {
				// extra param
			}*/
		} catch (NumberFormatException e){
			TestLog.e(AlgoParamHandler.class, e.toString() + " - " + e.getLocalizedMessage(), 5);
			return "Invalid parameters";
		}
		if (multiLine == true) {
			VMBridge bridge = PeaProperties.getVmBridge();
			return bridge.translate("key_derivation") + ": " + keyDerivation + "\n"
					+ bridge.translate("parameter")+ ": memory " + mem + ", time " + time 
					+ ((parallel.equals("")) ? "\n" : ", parallel " + parallel + "\n" )
					+ bridge.translate("cipher") + ": " + cipherAlgo + " in " + cipherMode + "\n"
					+ bridge.translate("hash") + ": " + hashAlgo + "\n\n"
					+ keyFile;
		} else {
			return keyDerivation + ", memory; " + mem + ", time; " + time 
					+ ((parallel.equals("")) ? ", " : ", Parallel parameter: " + parallel + ", " )
					+ cipherAlgo + " " + cipherMode + ", " + hashAlgo + ", " + keyFile;
		}
	}
	
	/**
	 * Get the size (= number of bytes) of the
	 * algorithm and parameter string
	 * 
	 * @return	size in number of bytes
	 */
	public static final int getAlgoParamStringSize() {
		return ALGO_PARAM_STRING_SIZE;
	}
	/**
	 * Get the default algorithm and parameter string
	 * from version 1
	 * 
	 * @return	old version string (version 1)
	 */
	public static final String getV1String() {
		return V1_ALGOPARAM_STRING;
	}
	/**
	 * Get the old algorithm and parameter string 
	 * as byte array
	 * 
	 * @return	old version string (version 1) as byte array
	 */
	public static final byte[] getV1Bytes() {
		return V1_ALGOPARAM_STRING.getBytes(Charset.forName("US-ASCII"));
	}

}
